@@ -12,7 +12,6 @@ This module covers the handshake TLS subprotocol, except for the key exchange
 mechanisms which are addressed with keyexchange.py.
 """
 
-from __future__ import absolute_import
 import math
 import os
 import struct
@@ -36,8 +35,7 @@ from scapy.fields import (
 )
 
 from scapy.compat import hex_bytes, orb, raw
-from scapy.config import conf
-from scapy.libs import six
+from scapy.config import conf, crypto_validator
 from scapy.packet import Packet, Raw, Padding
 from scapy.utils import randstring, repr_hex
 from scapy.layers.x509 import OCSP_Response
@@ -46,6 +44,7 @@ from scapy.layers.tls.basefields import (_tls_version, _TLSVersionField,
                                          _TLSClientVersionField)
 from scapy.layers.tls.extensions import (_ExtensionsLenField, _ExtensionsField,
                                          _cert_status_type,
+                                         TLS_Ext_PostHandshakeAuth,
                                          TLS_Ext_SupportedVersion_CH,
                                          TLS_Ext_SignatureAlgorithms,
                                          TLS_Ext_SupportedVersion_SH,
@@ -115,9 +114,15 @@ class _TLSHandshake(_GenericTLSSessionInheritance):
         """
         Covers both post_build- and post_dissection- context updates.
         """
-
-        self.tls_session.handshake_messages.append(msg_str)
-        self.tls_session.handshake_messages_parsed.append(self)
+        # RFC8446 sect 4.4.1
+        # "Note, however, that subsequent post-handshake authentications do not
+        # include each other, just the messages through the end of the main
+        # handshake."
+        if self.tls_session.post_handshake:
+            self.tls_session.post_handshake_messages.append(msg_str)
+        else:
+            self.tls_session.handshake_messages.append(msg_str)
+            self.tls_session.handshake_messages_parsed.append(self)
 
 
 ###############################################################################
@@ -180,7 +185,7 @@ class _CipherSuitesField(StrLenField):
         self.itemsize = struct.calcsize(itemfmt)
         i2s = self.i2s = {}
         s2i = self.s2i = {}
-        for k in six.iterkeys(dico):
+        for k in dico.keys():
             i2s[k] = dico[k]
             s2i[dico[k]] = k
 
@@ -338,9 +343,10 @@ class TLSClientHello(_TLSHandshake):
                             break
                     if s.sid:
                         s.middlebox_compatibility = True
-
                 if isinstance(e, TLS_Ext_SignatureAlgorithms):
                     s.advertised_sig_algs = e.sig_algs
+                if isinstance(e, TLS_Ext_PostHandshakeAuth):
+                    s.post_handshake_auth = True
 
 
 class TLS13ClientHello(_TLSHandshake):
@@ -398,7 +404,7 @@ class TLS13ClientHello(_TLSHandshake):
                         # For a resumed PSK, the hash function use
                         # to compute the binder must be the same
                         # as the one used to establish the original
-                        # conntection. For that, we assume that
+                        # connection. For that, we assume that
                         # the ciphersuite associate with the ticket
                         # is given as argument to tlsSession
                         # (see layers/tls/automaton_cli.py for an
@@ -465,10 +471,12 @@ class TLS13ClientHello(_TLSHandshake):
                         # RFC 8701: GREASE of TLS will send unknown versions
                         # here. We have to ignore them
                         if ver in _tls_version:
-                            self.tls_session.advertised_tls_version = ver
+                            s.advertised_tls_version = ver
                             break
                 if isinstance(e, TLS_Ext_SignatureAlgorithms):
                     s.advertised_sig_algs = e.sig_algs
+                if isinstance(e, TLS_Ext_PostHandshakeAuth):
+                    s.post_handshake_auth = True
 
 
 ###############################################################################
@@ -519,6 +527,11 @@ class TLSServerHello(_TLSHandshake):
                 return TLS13ServerHello
         return TLSServerHello
 
+    def build(self, *args, **kargs):
+        if self.getfieldval("sid") == b"" and self.tls_session:
+            self.sid = self.tls_session.sid
+        return super(TLSServerHello, self).build(*args, **kargs)
+
     def post_build(self, p, pay):
         if self.random_bytes is None:
             p = p[:10] + randstring(28) + p[10 + 28:]
@@ -667,7 +680,6 @@ class TLS13ServerHello(TLSServerHello):
             if not s.middlebox_compatibility:
                 s.triggered_pwcs_commit = True
         elif connection_end == "client":
-
             s.prcs = readConnState(ciphersuite=cs_cls,
                                    connection_end=connection_end,
                                    tls_version=s.tls_version)
@@ -700,12 +712,16 @@ class TLS13HelloRetryRequest(_TLSHandshake):
         fval = self.getfieldval("random_bytes")
         if fval is None:
             self.random_bytes = _tls_hello_retry_magic
+        if self.getfieldval("sid") == b"" and self.tls_session:
+            self.sid = self.tls_session.sid
         return _TLSHandshake.build(self)
 
+    @crypto_validator
     def tls_session_update(self, msg_str):
         s = self.tls_session
         s.tls13_retry = True
         s.tls13_client_pubshares = {}
+        # RFC8446 sect 4.4.1
         # If the server responds to a ClientHello with a HelloRetryRequest
         # The value of the first ClientHello is replaced by a message_hash
         if s.client_session_ticket:
@@ -802,7 +818,8 @@ class TLSEncryptedExtensions(_TLSHandshake):
                     s.wcs = self.tls_session.pwcs
                     s.triggered_pwcs_commit = False
                 else:
-                    s.triggered_prcs_commit = True
+                    s.triggered_pwcs_commit = True
+
 ###############################################################################
 #   Certificate                                                               #
 ###############################################################################
@@ -990,11 +1007,11 @@ class TLS13Certificate(_TLSHandshake):
         connection_end = self.tls_session.connection_end
         if connection_end == "client":
             if self.certs:
-                sc = [x.cert[1] for x in self.certs]
+                sc = [x.cert[1] for x in self.certs if hasattr(x, 'cert')]
                 self.tls_session.server_certs = sc
         else:
             if self.certs:
-                cc = [x.cert[1] for x in self.certs]
+                cc = [x.cert[1] for x in self.certs if hasattr(x, 'cert')]
                 self.tls_session.client_certs = cc
 
 
@@ -1181,6 +1198,11 @@ class TLS13CertificateRequest(_TLSHandshake):
                                     length_from=lambda pkt: pkt.msglen -
                                     pkt.cert_req_ctxt_len - 3)]
 
+    def tls_session_update(self, msg_str):
+        super(TLS13CertificateRequest, self).tls_session_update(msg_str)
+        self.tls_session.tls13_cert_req_ctxt = self.cert_req_ctxt
+
+
 ###############################################################################
 #   ServerHelloDone                                                           #
 ###############################################################################
@@ -1203,11 +1225,16 @@ class TLSCertificateVerify(_TLSHandshake):
                    _TLSSignatureField("sig", None,
                                       length_from=lambda pkt: pkt.msglen)]
 
+    # See https://datatracker.ietf.org/doc/html/rfc8446#section-4.4 for how to compute
+    # the signature.
+
     def build(self, *args, **kargs):
         sig = self.getfieldval("sig")
         if sig is None:
             s = self.tls_session
             m = b"".join(s.handshake_messages)
+            if s.post_handshake:
+                m += b"".join(s.post_handshake_messages)
             tls_version = s.tls_version
             if tls_version is None:
                 tls_version = s.advertised_tls_version
@@ -1228,6 +1255,8 @@ class TLSCertificateVerify(_TLSHandshake):
     def post_dissection(self, pkt):
         s = self.tls_session
         m = b"".join(s.handshake_messages)
+        if s.post_handshake:
+            m += b"".join(s.post_handshake_messages)
         tls_version = s.tls_version
         if tls_version is None:
             tls_version = s.advertised_tls_version
@@ -1331,7 +1360,14 @@ class TLSClientKeyExchange(_TLSHandshake):
                 self.tls_session.session_hash = (
                     Hash_MD5().digest(to_hash) + Hash_SHA().digest(to_hash)
                 )
-            self.tls_session.compute_ms_and_derive_keys()
+            if self.tls_session.pre_master_secret:
+                self.tls_session.compute_ms_and_derive_keys()
+
+        if not self.tls_session.master_secret:
+            # There are still no master secret (we're just passive)
+            if self.tls_session.use_nss_master_secret_if_present():
+                # we have a NSS file
+                self.tls_session.compute_ms_and_derive_keys()
 
 
 ###############################################################################
@@ -1342,7 +1378,7 @@ class _VerifyDataField(StrLenField):
     def getfield(self, pkt, s):
         if pkt.tls_session.tls_version == 0x0300:
             sep = 36
-        elif pkt.tls_session.tls_version >= 0x0304:
+        elif pkt.tls_session.tls_version and pkt.tls_session.tls_version >= 0x0304:
             sep = pkt.tls_session.rcs.hash.hash_len
         else:
             sep = 12
@@ -1360,6 +1396,8 @@ class TLSFinished(_TLSHandshake):
         if fval is None:
             s = self.tls_session
             handshake_msg = b"".join(s.handshake_messages)
+            if s.post_handshake:
+                handshake_msg += b"".join(s.post_handshake_messages)
             con_end = s.connection_end
             tls_version = s.tls_version
             if tls_version is None:
@@ -1369,13 +1407,16 @@ class TLSFinished(_TLSHandshake):
                 self.vdata = s.wcs.prf.compute_verify_data(con_end, "write",
                                                            handshake_msg, ms)
             else:
-                self.vdata = s.compute_tls13_verify_data(con_end, "write")
+                self.vdata = s.compute_tls13_verify_data(con_end, "write",
+                                                         handshake_msg)
         return _TLSHandshake.build(self, *args, **kargs)
 
     def post_dissection(self, pkt):
         s = self.tls_session
         if not s.frozen:
             handshake_msg = b"".join(s.handshake_messages)
+            if s.post_handshake:
+                handshake_msg += b"".join(s.post_handshake_messages)
             tls_version = s.tls_version
             if tls_version is None:
                 tls_version = s.advertised_tls_version
@@ -1389,7 +1430,8 @@ class TLSFinished(_TLSHandshake):
                     log_runtime.info("TLS: invalid Finished received [%s]", pkt_info)  # noqa: E501
             elif tls_version >= 0x0304:
                 con_end = s.connection_end
-                verify_data = s.compute_tls13_verify_data(con_end, "read")
+                verify_data = s.compute_tls13_verify_data(con_end, "read",
+                                                          handshake_msg)
                 if self.vdata != verify_data:
                     pkt_info = pkt.firstlayer().summary()
                     log_runtime.info("TLS: invalid Finished received [%s]", pkt_info)  # noqa: E501
@@ -1400,7 +1442,7 @@ class TLSFinished(_TLSHandshake):
         tls_version = s.tls_version
         if tls_version is None:
             tls_version = s.advertised_tls_version
-        if tls_version >= 0x0304:
+        if tls_version >= 0x0304 and not s.post_handshake:
             s.pwcs = writeConnState(ciphersuite=type(s.wcs.ciphersuite),
                                     connection_end=s.connection_end,
                                     tls_version=s.tls_version)
@@ -1410,6 +1452,9 @@ class TLSFinished(_TLSHandshake):
             elif s.connection_end == "client":
                 s.compute_tls13_traffic_secrets_end()
                 s.compute_tls13_resumption_secret()
+        if s.connection_end == "client":
+            s.post_handshake = True
+            s.post_handshake_messages = []
 
     def post_dissection_tls_session_update(self, msg_str):
         self.tls_session_update(msg_str)
@@ -1417,7 +1462,7 @@ class TLSFinished(_TLSHandshake):
         tls_version = s.tls_version
         if tls_version is None:
             tls_version = s.advertised_tls_version
-        if tls_version >= 0x0304:
+        if tls_version >= 0x0304 and not s.post_handshake:
             s.prcs = readConnState(ciphersuite=type(s.rcs.ciphersuite),
                                    connection_end=s.connection_end,
                                    tls_version=s.tls_version)
@@ -1427,6 +1472,9 @@ class TLSFinished(_TLSHandshake):
             elif s.connection_end == "server":
                 s.compute_tls13_traffic_secrets_end()
                 s.compute_tls13_resumption_secret()
+        if s.connection_end == "server":
+            s.post_handshake = True
+            s.post_handshake_messages = []
 
 
 # Additional handshake messages
@@ -1654,7 +1702,6 @@ class TLS13NewSessionTicket(_TLSHandshake):
         return _TLSHandshake.build(self)
 
     def post_dissection_tls_session_update(self, msg_str):
-        self.tls_session_update(msg_str)
         if self.tls_session.connection_end == "client":
             self.tls_session.client_session_ticket = self.ticket
 
